;; Generated package description from tp.el  -*- no-byte-compile: t -*-
(define-package "tp" "0.7.0.20250206.81229" "Utilities for transient menus that POST to an API" '((emacs "28.1") (transient "0.5.0")) :commit "cce2dfe0ec2b5c070cb13a7bdf95695eeb6e3caf" :authors '(("Marty Hiatt" . "mousebot@disroot.org")) :maintainer '("Marty Hiatt" . "mousebot@disroot.org") :keywords '("convenience" "api" "requests") :url "https://codeberg.org/martianh/tp.el")
